// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:matrix_sdk/src/event/event.dart';
import 'package:matrix_sdk/src/event/room/message_event.dart';
import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/event/room/state/state_event.dart';
import 'package:matrix_sdk/src/room/room.dart';
import 'package:matrix_sdk/src/storables.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/store.dart';
import 'package:meta/meta.dart';

/// Events map that will lazily load from the given
/// [store].
class Events extends Storables<EventId, Event> {

  final Room _room;

  Events(this._room, {Store store}) : super(store: store);

  @override
  @protected
  Future<void> saveAll(Iterable<Event> events)
    => store?.addEvents(_room, events);

  @override
  @protected
  Future<Event> retrieve(EventId id) => store?.getEvent(_room, id);

  @override
  @protected
  Stream<Event> retrieveAll() => store?.getEvents(_room);

  @override
  @protected
  Stream<Event> retrieveUpTo(int count, {@required Event from})
    => store?.getEvents(_room, count: count, from: from as RoomEvent);

  @protected
  @override
  Stream<Event> retrieveRemote(int count, {@required Event from})
    => _room.retrieveRemoteEvents(count);

  @protected
  @override
  Future<void> removeInStore(EventId id) async {
    await store?.removeEvent(id);
  }

  @override
  Future<void> addAll(Iterable<Event> values) async {

    // Remove echo events
    for (final event in values) {
      // TODO: Optimize?
      final localDupe = await all().firstWhere((e) {
        if (e.unsigned != null && e.unsigned.transactionId != null) {
          return e.unsigned.transactionId == event.unsigned.transactionId;
        } else {
          return false;
        }
      }, orElse: () => null);
      if (localDupe != null) {
        await remove(localDupe.id);
      }
    }

    await super.addAll(values);
  }

  @override
  Iterable<Event> sort(Iterable<Event> items) {
    final list = items.toList(growable: false);
    list.sort((a, b) {
      if (a is RoomEvent && b is RoomEvent) {
        return a.time.compareTo(b.time);
      } else {
        return a.id.toString().compareTo(b.id.toString());
      }
    });

    return list.reversed;
  }

  Stream<MessageEvent> messages() =>
      all().where((event) => event is MessageEvent).cast();

  Stream<StateEvent> state() =>
      all().where((event) => event is StateEvent).cast();
}
