// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:matrix_sdk/src/event/event.dart';
import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/room/room.dart';
import 'package:matrix_sdk/src/user.dart';
import 'package:meta/meta.dart';

const String _messageEventType = 'm.room.message';

abstract class MessageEventContent extends EventContent {

  @override
  final String matrixType = _messageEventType;

  @protected
  String get messageType;

  @override
  Map<String, dynamic> toJson() => {
    'msgtype': messageType
  };
}

abstract class MessageEvent extends RoomEvent {

  static const String type = _messageEventType;

  @override
  final String matrixType = _messageEventType;

  @override
  final MessageEventContent content;

  MessageEvent(
    this.content,
    EventId id,
    Room room,
    User sender,
    DateTime time,
    UnsignedData unsigned,
    {SentState sentState = SentState.sent})
    : super(id, room, sender, time, unsigned, sentState: sentState);

  factory MessageEvent.fromJson(
      EventId id, Room room, User sender,
      DateTime time, UnsignedData unsigned, Map<String, dynamic> json) {

    var msgtype = json['content']['msgtype'];

    switch (msgtype) {
      case ImageMessage.type:
        final body = json['content']['body'];

        var url = json['content']['url'];
        if (url == null) {
          return null;
        }

        var info = json['content']['info'];
        if (info != null) {
          info = ImageInfo(
            width: json['content']['info']['w'],
            height: json['content']['info']['h']
          );
        }

        url = Uri.parse(url);
        return ImageMessageEvent(
          ImageMessage(body: body, url: url, info: info),
          id,
          room,
          sender,
          time,
          unsigned);
      case TextMessage.type:
      default:
        final body = json['content']['body'];
        final formattedBody = json['content']['formatted_body'];
        return TextMessageEvent(
          TextMessage(body: body, formattedBody: formattedBody), id, room, sender, time, unsigned);
    }
  }
}

class TextMessage extends MessageEventContent {
  static const String type = 'm.text';
  @override
  final String messageType = type;

  final String body;
  final String formattedBody;

  TextMessage({@required this.body, String formattedBody})
    : formattedBody = formattedBody ?? body;

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'body': body,
      'formatted_body': formattedBody
    });
}

class TextMessageEvent extends MessageEvent {

  @override
  final TextMessage content;

  TextMessageEvent(
    this.content,
    EventId id,
    Room room,
    User sender,
    DateTime time,
    UnsignedData unsigned,
    {SentState sentState = SentState.sent})
    : super(content, id, room, sender, time, unsigned, sentState: sentState);
}

class ImageMessage extends MessageEventContent {
  static const String type = 'm.image';
  @override
  final String messageType = type;

  final String body;
  final Uri url;
  final ImageInfo info;

  ImageMessage({@required this.body, this.url, this.info});

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'body': body,
      'url': url.toString(),
      'info': {
        'w': info.width,
        'h': info.height
      }
    });
}

class ImageInfo {
  final int width;
  final int height;

  ImageInfo({@required this.width, @required this.height});
}

class ImageMessageEvent extends MessageEvent {

  @override
  final ImageMessage content;

  ImageMessageEvent(
      this.content,
      EventId id,
      Room room,
      User sender,
      DateTime time,
      UnsignedData unsigned,
      {SentState sentState = SentState.sent})
      : super(content, id, room, sender, time, unsigned, sentState: sentState);
}