// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'dart:convert';
import 'dart:typed_data';

import 'package:chopper/chopper.dart';
import 'package:matrix_sdk/src/api/client.dart';
import 'package:matrix_sdk/src/api/media.dart';
import 'package:matrix_sdk/src/exception.dart';

import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/store.dart';
import 'package:matrix_sdk/src/user.dart';
import 'package:matrix_sdk/src/util/chopper.dart';

import 'package:kiwi/kiwi.dart';

/// Represents a Matrix homeserver. Also used as the main entry point
/// of the SDK.
class Homeserver {
  final ChopperClient _chopper;
  // TODO: Make lazy
  ClientService get _client => _chopper.getService();
  MediaService get _media => _chopper.getService();

  final Uri uri;

  /// Dependency injection scope.
  /// @internal
  String scope;

  Homeserver(this.uri)
      : _chopper = createChopper(uri: uri){

    scope = uri.toString() + hashCode.toString();

    Container().registerInstance(this, name: scope);
    Container().registerInstance(_client, name: scope);
    Container().registerInstance(_media, name: scope);
  }

  /// Logs in to this homeserver and returns a [LocalUser], which can be
  /// used for further operations.
  ///
  /// [user]: Can be either a [UserId] or just a [Username].
  /// [password]: Password of the given [user].
  /// [store]: Store where all user data (rooms, events, etc) will be stored.
  Future<LocalUser> login(UserIdentifier user,
                          String password,
                          {Store store}) async {
    final response = await _client.login(json.encode({
      'type': 'm.login.password',
      'identifier': user.toIdentifierJson(),
      'password': password
    }));

    final body = json.decode(response.body);

    String accessToken = body['access_token'];
    final userId = UserId(body['user_id']);

    final localUser = LocalUser(
      id: userId,
      store: store,
      homeserver: this,
      accessToken: accessToken
    );

    await store?.open(user: localUser);

    await store?.setUserId(userId);
    await store?.setHomeserver(this);
    await store?.setAccessToken(accessToken);

    return localUser;
  }

  /// Checks whether the given [username] is available for
  /// registration on this [Homeserver].
  Future<bool> isUsernameAvailable(Username username) async {
    var response;

    try {
      response = await _client.registerAvailable(username.toString());
    } on UsernameInUseException {
      return false;
    }

    switch (response.statusCode) {
      case 200:
        return true;
      default:
        return false;
    }
  }

  /// Download content via this [Homeserver].
  Future<Uint8List> download(Uri uri) async {
    if (uri == null) {
      throw ArgumentError.value(uri, 'uri', 'uri must not be null');
    }

    if (uri.scheme != 'mxc') {
      throw ArgumentError.value(uri, 'uri', 'uri must be an mxc uri');
    }

    var server = uri.authority;
    var mediaId = uri.pathSegments[0];

    final response = await _media.download(server, mediaId);
    return response.bodyBytes;
  }
}
