// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'dart:convert';

import 'package:kiwi/kiwi.dart';
import 'package:matrix_sdk/src/api/client.dart';
import 'package:matrix_sdk/src/event/event.dart';
import 'package:matrix_sdk/src/event/events.dart';
import 'package:matrix_sdk/src/event/room/message_event.dart';
import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/event/room/state/avatar_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/member_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/name_change_event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/members.dart';
import 'package:matrix_sdk/src/store.dart';
import 'package:matrix_sdk/src/user.dart';
import 'package:matrix_sdk/src/util/json.dart';
import 'package:chopper/chopper.dart';
import 'package:matrix_sdk/src/util/random.dart';
import 'package:meta/meta.dart';

class Room extends Identifiable<RoomId> implements JsonWritable {
  @override
  final RoomId id;

  Events _events;
  Events get events => _events;
  final Members members;

  //// @internal
  String previousBatch;

  final String _scope;
  ClientService _client;

  Room(this.id,
      {@required String scope,
        Store store,
        String name,
        Uri avatarUrl,
        String previousBatch})
      : _scope = scope,
        _name = name,
        _avatarUrl = avatarUrl,
        members = Members(id, store: store) {
    _events = Events(this, store: store);

    _client = Container().resolve(_scope);
  }

  String _name;
  String get name => _name ?? id.toString();

  Uri _avatarUrl;
  Uri get avatarUrl => _avatarUrl;

  factory Room.fromJson(RoomId id, Map<String, dynamic> json) {
    final String name = json['name'];
    final Uri avatarUrl = Uri.parse(json['avatar_url']);
    return Room(id, name: name, avatarUrl: avatarUrl);
  }

  // TODO: Generate
  dynamic toJson() => { 'name': name, 'avatar_url': avatarUrl };

  /// @internal
  Stream<Event> retrieveRemoteEvents(int count) async* {
    final response = await _client.roomMessages(
      roomId: id.toString(),
      limit: count,
      from: previousBatch,
    );

    final body = json.decode(response.body);

    previousBatch = body['end'];
    
    final processed = await _processTimeline(body['chunk']);
    for (final event in processed) {
      // TODO: When most even types are supported, don't check if they're
      // null anymore (null is now returned if the event type is unsupported)
      if (event != null) {
        yield event;
      }
    }
  }

  /// @internal
  Future<void> processSync(Map<String, dynamic> response) async {
    // Only set previous batch from sync when none is set
    if (previousBatch == null) {
      previousBatch = response['timeline']['prev_batch'];
    }


    await _processState(response['state']['events']);

    final processed = await _processTimeline(response['timeline']['events']);
    await events.addAll(processed);
  }

  Future<void> _processState(List<dynamic> rawStateEvents) async {

    // Map to assign to the events
    final users = List<User>();

    final convertedEvents = rawStateEvents
      .map((raw) async => await Event.fromJson(this, raw));

    (await Future.wait(convertedEvents))
      .forEach((event) async {
        if (event is JoinEvent) {
          users.add(User(
            id: event.content.subjectId,
            name: event.content.displayName
          ));
        }

        if (event is NameChangeEvent) {
          _name = event.content.name;
        }

        if (event is AvatarChangeEvent) {
          _avatarUrl = event.content.url;
        }
      });

    await members.addAll(users);
  }

  Future<List<Event>> _processTimeline(List<dynamic> rawEvents) async
    => await Future.wait(rawEvents.map((raw) => Event.fromJson(this, raw)));
}

class JoinedRoom extends Room {
  JoinedRoom(
      RoomId id,
      {String scope,
       Store store,
       String name,
       Uri avatarUrl,
       String previousBatch})
      : super(id,
        scope: scope,
        name: name,
        store: store,
        avatarUrl:
        avatarUrl,
        previousBatch: previousBatch
      );

  factory JoinedRoom.fromJson(
      RoomId id, String scope, Store store, Map<String, dynamic> json) {
    final String name = json['name'];
    return JoinedRoom(id,
      name: name,
      store: store,
      scope: scope
    );
  }

  /// Send an [Event] to the [Room].
  /// Returns the up to date [SentState] of the sent event. (Not
  /// including [SentState.received].
  Stream<SentState> send(EventContent content) async* {
    final txnId = randomString();
    final event = Event.fromContent(content,
      id: EventId(randomString()),
      room: this,
      time: DateTime.now(),
      sender: Container().resolve<LocalUser>(_scope),
      unsigned: UnsignedData(transactionId: txnId),
      sentState: SentState.waiting
    );

    await events.addAll([event]);
    yield SentState.waiting;

    final response = await _client.send(
      roomId: id.toString(),
      eventType: content.matrixType,
      txnId: txnId,
      content: json.encode(content.toJson()));

    final body = json.decode(response.body);
    final eventId = EventId(body['event_id']);

    final sentEvent = Event.fromContent(content,
      id: eventId,
      room: this,
      time: DateTime.now(),
      sender: Container().resolve<LocalUser>(_scope),
      unsigned: UnsignedData(transactionId: txnId),
      sentState: SentState.sent
    );

    await events.addAll([sentEvent]);
    yield SentState.sent;
  }
}
