// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'dart:convert';

import 'package:chopper/chopper.dart';
import 'package:kiwi/kiwi.dart';
import 'package:matrix_sdk/src/api/access_token_interceptor.dart';
import 'package:matrix_sdk/src/api/client.dart';
import 'package:matrix_sdk/src/homeserver.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/room/room.dart';
import 'package:matrix_sdk/src/room/rooms.dart';
import 'package:matrix_sdk/src/store.dart';
import 'package:matrix_sdk/src/util/chopper.dart';
import 'package:meta/meta.dart';

class User extends Identifiable<UserId> {
  @override
  final UserId id;

  String _name;
  String get name => _name ?? id.toString();

  User({@required this.id, String name}) : _name = name;
}

/// A user which is authenticated and can send messages, join rooms etc.
class LocalUser extends User {
  final Store _store;

  final Homeserver homeserver;
  final String accessToken;

  bool _isLoggedOut = false;
  bool get isLoggedOut => _isLoggedOut;

  /// @internal
  final String scope;

  String _syncToken;

  final ChopperClient _chopper;
  ClientService get _client => _chopper.getService();

  static Future<LocalUser> fromStore(Store store) async {
    final isOpen = await store.open();

    if (!isOpen) {
      return null;
    }

    final userId = await store.getUserId();
    if (userId == null) {
      return null;
    }

    final hs = await store.getHomeserver();
    if (hs == null) {
      return null;
    }

    final accessToken = await store.getAccessToken();
    if (accessToken == null) {
      return null;
    }

    return LocalUser(
      id: userId,
      homeserver: hs,
      accessToken: accessToken,
      store: store
    );
  }

  LocalUser(
    {@required UserId id,
    @required this.homeserver,
    @required this.accessToken,
    @required Store store})
    : _chopper = createChopper(
          uri: homeserver.uri,
          interceptors: [AccessTokenInterceptor(accessToken)]),
      rooms = Rooms(store: store),
      _store = store,
      scope = homeserver.scope + id.toString(),
      super(id: id) {
    Container().registerInstance(this, name: scope);
    Container().registerInstance(_client, name: scope);
  }

  final Rooms rooms;

  /// Invalidates the access token of the user. Makes all
  /// [LocalUser] calls unusable.
  Future<void> logout() async {
    await _client.logout();
    _isLoggedOut = true;
  }

  // Sync //
  /// Syncs data with the user's [homeserver]. Use the returned [Stream] to
  /// get when the sync has been successful (`true`) or unsuccessful (`false`).
  Stream<bool> sync() async* {
    if (isLoggedOut) {
      throw StateError("The user can not be logged out");
    }

    _syncToken = await _store?.getSyncToken() ?? _syncToken;

    while (true) {
      final Response response = await _client.sync(
        since: _syncToken,
        fullState: true,
        filter: json.encode(const {
          'room': {
            'state': {
              'lazy_load_members': true
            }
          }
        })
      );
      final body = json.decode(response.body);

      // Store sync token
      _syncToken = body['next_batch'] as String;
      await _store?.setSyncToken(_syncToken);

      await _processRooms(body);

      yield true;

      await Future.delayed(Duration(seconds: 5));
    }
  }

  Future<void> _processRooms(Map<String, dynamic> body) async {
    await _processJoinedRooms(body['rooms']['join']);
  }

  Future<void> _processJoinedRooms(Map<String, dynamic> joinedRooms) async {
    final processed = List<JoinedRoom>();

    for (final entry in joinedRooms.entries) {
      final roomId = RoomId(entry.key);
      final j = entry.value;

      Room room = await rooms[roomId];
      if (room == null) {
        room = JoinedRoom.fromJson(roomId, scope, _store, j);
      }

      await room.processSync(j);

      processed.add(room);
    }

    // Save rooms
    await rooms.addAll(processed);
  }

  // End Sync //
}
