// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:matrix_sdk/matrix_sdk.dart';
import "package:test/test.dart";

import 'package:dotenv/dotenv.dart';

void main() {

  group("Identifier validity", () {
    test("UserId", () {
      assert(Username.isValid("michael-scott"));
      assert(Username.isValid("pit_pattle"));
      assert(Username.isValid("joe"));

      assert(!Username.isValid("#\$^%\$43#\$^"));
      assert(!Username.isValid("wil@ko"));
      assert(!Username.isValid("@matthew"));
    });

    test("Username", () {
      assert(!UserId.isValidFullyQualified("#\$^%\$43#\$^"));
      assert(!UserId.isValidFullyQualified("@pit"));
      assert(!UserId.isValidFullyQualified("pit"));
      assert(!UserId.isValidFullyQualified("pi&*3t"));
      assert(!UserId.isValidFullyQualified("pi&*3:t.tk"));

      assert(UserId.isValidFullyQualified("@pit:pattle.im"));
      assert(UserId.isValidFullyQualified("@joe:matrix.org"));
      assert(UserId.isValidFullyQualified("@jim:192.168.0.1"));
      assert(UserId.isValidFullyQualified("@jim:192.168.0.1:9043"));
    });
  });

  // The intergration tests will use the provided username and password from
  // the .env file, and assume that the given user is a member of the
  // #dart-sdk-test:matrix.org room.
  group("Integration", () {
    Homeserver homeserver;

    LocalUser user;

    setUp(() async {
      homeserver = Homeserver(
        Uri.parse("https://matrix.org")
      );

      // Load .env
      load();
    });

    test('Login', () async {
      user = await homeserver.login(
        Username(env['username']),
        env['password']
      );
    });

    test('Test room exists', () async {

      await for(bool success in user.sync()) {

        assert(success);

        final room = await user.rooms.all().firstWhere(
          (room) => room.id == RoomId('!otDeInmGHULHzwJyod:matrix.org'),
          orElse: () => null) as JoinedRoom;

        assert(room != null);

        var message = await room.events.upTo(20)
          .firstWhere((Event event) =>
            event is TextMessageEvent &&
            event.sender.id == UserId('@wilko:matrix.org') &&
            event.sender.name == 'Wilko');

        assert(message != null);

        var wilko = await room.members.all()
          .firstWhere((user) => user.name == 'Wilko');

        assert(wilko != null);

        final shipment = room.send(TextMessage(body: 'Sent from the Dart Matrix SDK'));
        await for (final sentState in shipment) {
          final sentEvent = await room.events.upTo(1).first as TextMessageEvent;
          assert(sentEvent.content.body == 'Sent from the Dart Matrix SDK');
        }

        // There seems to be a bug in Matrix
        // where the avatar state event is not given for this room
        //var avatarBytes = await homeserver.download(room.avatarUrl);
        //assert(avatarBytes.length == 95245);

        final imageMessage = await room.events.all()
            .firstWhere((event) => event is ImageMessageEvent) as ImageMessageEvent;

        final imageBytes = await homeserver.download(imageMessage.content.url);
        assert(imageBytes.length == 155471);

        assert(imageMessage.content.info.width == 720);
        assert(imageMessage.content.info.height == 1280);

        break;
      }

      await user.logout();
      assert(user.isLoggedOut);
    }, timeout: const Timeout(const Duration(minutes: 5)));
  });
}
